#!/bin/python
#-----------------------------------------------------------------------------
# File Name : 
# Author: Emre Neftci
#
# Creation Date : Tue Nov  5 16:26:06 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
import numpy as np
import pandas as pd
import torch, bisect
from torchvision.transforms import Compose,ToTensor,Normalize,Lambda

def find_first(a, tgt):
    return bisect.bisect_left(a, tgt)

class toOneHot(object):
    def __init__(self, num_classes):
        self.num_classes = num_classes

    def __call__(self, integers):
        y_onehot = torch.FloatTensor(integers.shape[0], self.num_classes)
        y_onehot.zero_()
        return y_onehot.scatter_(1, torch.LongTensor(integers), 1)

class Downsample(object):
    """Resize the address event Tensor to the given size.

    Args:
        factor: : Desired resize factor. Applied to all dimensions including time
    """
    def __init__(self, factor):
        assert isinstance(factor, int) or hasattr(factor, '__iter__')
        self.factor = factor

    def __call__(self, tmad):
        return tmad//self.factor

    def __repr__(self):
        return self.__class__.__name__ + '()'

class Crop(object):
    def __init__(self, low_crop, high_crop):
        '''
        Crop all dimensions
        '''
        self.low = low_crop
        self.high = high_crop

    def __call__(self, tmad):
        idx = np.where(np.any(tmad>high_crop, axis=1))
        tmad = np.delete(tmad,idx,0)
        idx = np.where(np.any(tmad<high_crop, axis=1))
        tmad = np.delete(tmad,idx,0)
        return tmad

    def __repr__(self):
        return self.__class__.__name__ + '()'

class CropDims(object):
    def __init__(self, low_crop, high_crop, dims):
        self.low_crop = low_crop
        self.high_crop = high_crop
        self.dims = dims

    def __call__(self, tmad):
        for i, d in enumerate(self.dims):
            idx = np.where(tmad[:,d]>=self.high_crop[i])
            tmad = np.delete(tmad,idx,0)
            idx = np.where(tmad[:,d]<self.low_crop[i])
            tmad = np.delete(tmad,idx,0)
            #Normalize
            tmad[:,d]=tmad[:,d]-self.low_crop[i]
        return tmad

    def __repr__(self):
        return self.__class__.__name__ + '()'

class Attention(object):
    def __init__(self, n_attention_events, size):
        '''
        Crop around the median event in the last n_events.
        '''
        self.att_shape = np.array(size[1:], dtype=np.int64)
        self.n_att_events = n_attention_events

    def __call__(self, tmad):
        df = pd.DataFrame(tmad, columns=['t', 'p', 'x', 'y'])
        # compute centroid in x and y
        centroids = df.loc[:, ['x', 'y']].rolling(window=self.n_att_events,
                                                  min_periods=1).median().astype(int)
        # re-address (translate) events with respect to centroid corner
        df.loc[:, ['x', 'y']] -= centroids - self.att_shape // 2
        # remove out of range events
        df = df.loc[(df.x >= 0) & (df.x < self.att_shape[1]) & (df.y >= 0) & (df.y < self.att_shape[0])]
        return df.to_numpy()

    def __repr__(self):
        return self.__class__.__name__ + '()'

class ToChannelHeightWidth(object):
    def __call__(self, tmad):
        n = tmad.shape[1]
        if n==2:
            o = np.zeros(tmad.shape[0], dtype=tmad.dtype)
            return np.column_stack([tmad, o, o])
            
        elif n==4:
            return tmad

        else:
            raise TypeError('Wrong number of dimensions. Found {0}, expected 1 or 3'.format(n-1))

    def __repr__(self):
        return self.__class__.__name__ + '()'

class ToCountFrame(object):
    """Convert Address Events to Binary tensor.

    Converts a numpy.ndarray (T x H x W x C) to a torch.FloatTensor of shape (T x C x H x W) in the range [0., 1., ...] 
    """
    def __init__(self, T=500, size=[2, 32, 32]):
        self.T = T
        self.size = size

    def __call__(self, tmad):
        times = tmad[:,0]
        t_start = times[0]
        t_end = times[-1]
        addrs = tmad[:,1:]

        ts = range(0, self.T)
        chunks = np.zeros([len(ts)] + self.size, dtype='int8')
        idx_start = 0
        idx_end = 0
        for i, t in enumerate(ts):
            idx_end += find_first(times[idx_end:], t)
            if idx_end > idx_start:
                ee = addrs[idx_start:idx_end]
                i_pol_x_y = (i, ee[:, 0], ee[:, 1], ee[:, 2])
                np.add.at(chunks, i_pol_x_y, 1)
            idx_start = idx_end
        return chunks

    def __repr__(self):
        return self.__class__.__name__ + '()'

class Repeat(object):
    '''
    Replicate np.array (C) as (n_repeat X C). This is useful to transform sample labels into sequences
    '''
    def __init__(self, n_repeat):
        self.n_repeat = n_repeat

    def __call__(self, target):
        return np.tile(np.expand_dims(target,0),[self.n_repeat,1])

    def __repr__(self):
        return self.__class__.__name__ + '()'

class ToTensor(object):
    """Convert a ``numpy.ndarray`` to tensor.

    Converts a numpy.ndarray (T x H x W x C) to a torch.FloatTensor of shape (T X H x W x C)
    """

    def __call__(self, frame):
        """
        Args:
            frame (numpy.ndarray): numpy array of frames

        Returns:
            Tensor: Converted data.
        """
        return torch.FloatTensor(frame)

    def __repr__(self):
        return self.__class__.__name__ + '()'






#!/bin/python
#-----------------------------------------------------------------------------
# File Name : create_hdf5.py
# Author: Emre Neftci
#
# Creation Date : Tue Nov  5 13:15:54 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
import numpy as np
from tqdm import tqdm
import scipy.misc
import h5py
import glob
import torch.utils.data
from ..events_timeslices import *
from ..utils import *
import os

def create_events_hdf5(directory, hdf5_filename):
    fns_train = gather_aedat(directory,1,24)
    fns_test = gather_aedat (directory,24,30)
    test_keys = []
    train_keys = []

    with h5py.File(hdf5_filename, 'w') as f:
        f.clear()

        key = 0
        metas = []
        data_grp = f.create_group('data')
        extra_grp = f.create_group('extra')
        for file_d in tqdm(fns_train+fns_test):
            istrain = file_d in fns_train
            data, labels_starttime = aedat_to_events(file_d)
            tms = data[:,0]
            ads = data[:,1:]
            lbls = labels_starttime[:,0]
            start_tms = labels_starttime[:,1]
            end_tms = labels_starttime[:,2]
            out = []

            for i, v in enumerate(lbls):
                if istrain: 
                    train_keys.append(key)
                else:
                    test_keys.append(key)
                s_ = get_slice(tms, ads, start_tms[i], end_tms[i])
                times = s_[0]
                addrs = s_[1]
                subj, light = file_d.split('/')[-1].split('.')[0].split('_')[:2]
                metas.append({'key':str(key), 'subject':subj,'light condition':light, 'training sample':istrain}) 
                subgrp = data_grp.create_group(str(key))
                tm_dset = subgrp.create_dataset('times' , data=times, dtype=np.uint32)
                ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype=np.uint8)
                lbl_dset= subgrp.create_dataset('labels', data=lbls[i]-1, dtype=np.uint8)
                subgrp.attrs['meta_info']= str(metas[-1])
                assert lbls[i]-1 in range(11)
                key += 1
        extra_grp.create_dataset('train_keys', data=train_keys)
        extra_grp.create_dataset('test_keys', data=test_keys)
        extra_grp.attrs['N'] = len(train_keys) + len(test_keys)
        extra_grp.attrs['Ntrain'] = len(train_keys)
        extra_grp.attrs['Ntest'] = len(test_keys)
            
def gather_aedat(directory, start_id, end_id, filename_prefix = 'user'):
    if not os.path.isdir(directory):
        raise FileNotFoundError("DVS Gestures Dataset not found, looked at: {}".format(directory))
    import glob
    fns = []
    for i in range(start_id,end_id):
        search_mask = directory+'/'+filename_prefix+"{0:02d}".format(i)+'*.aedat'
        glob_out = glob.glob(search_mask)
        if len(glob_out)>0:
            fns+=glob_out
    return fns





#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Author: Emre Neftci
#
# Creation Date : Fri 01 Dec 2017 10:05:17 PM PST
# Last Modified : Sun 29 Jul 2018 01:39:06 PM PDT
#
# Copyright : (c)
# Licence : Apache License, Version 2.0
#-----------------------------------------------------------------------------
import struct
import time
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from ..neuromorphic_dataset import NeuromorphicDataset 
from ..events_timeslices import *
from ..transforms import *
from .create_hdf5 import create_events_hdf5
import os

mapping = { 0 :'0',
            1 :'1',
            2 :'2',
            3 :'3',
            4 :'4',
            5 :'5',
            6 :'6',
            7 :'7',
            8 :'8',
            9 :'9'}

class NMNISTDataset(NeuromorphicDataset):
    resources_url = [['https://www.dropbox.com/sh/tg2ljlbmtzygrag/AABlMOuR15ugeOxMCX0Pvoxga/Train.zip?dl=1',None, 'Train.zip'],
                     ['https://www.dropbox.com/sh/tg2ljlbmtzygrag/AADSKgJ2CjaBWh75HnTNZyhca/Test.zip?dl=1', None, 'Test.zip']]
    directory = 'data/nmnist/'
    resources_local = [directory+'Train', directory+'Test']

    def __init__(
            self, 
            root,
            train=True,
            transform=None,
            target_transform=None,
            download_and_create=True,
            chunk_size = 500):

        self.n = 0
        self.nclasses = 10
        self.download_and_create = download_and_create
        self.root = root
        self.train = train 
        self.chunk_size = chunk_size

        super(NMNISTDataset, self).__init__(
                root,
                transform=transform,
                target_transform=target_transform )
        
        with h5py.File(root, 'r', swmr=True, libver="latest") as f:
            if train:
                self.n = f['extra'].attrs['Ntrain']
                self.keys = f['extra']['train_keys']
            else:
                self.n = f['extra'].attrs['Ntest']
                self.keys = f['extra']['test_keys']

    def download(self):
        isexisting = super(NMNISTDataset, self).download()

    def create_hdf5(self):
        create_events_hdf5(self.directory, self.root)


    def __len__(self):
        return self.n
        
    def __getitem__(self, key):
        #Important to open and close in getitem to enable num_workers>0
        with h5py.File(self.root, 'r', swmr=True, libver="latest") as f:
            if not self.train:
                key = key + f['extra'].attrs['Ntrain']
            data, target = sample(
                    f,
                    key,
                    T = self.chunk_size,
                    shuffle=self.train)

        if self.transform is not None:
            data = self.transform(data)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return data, target

def sample(hdf5_file,
        key,
        T = 500,
        shuffle = False):
    dset = hdf5_file['data'][str(key)]
    label = dset['labels'][()]
    tend = dset['times'][-1] 
    start_time = 0

    tmad = get_tmad_slice(dset['times'][()], dset['addrs'][()], start_time, T*1000)
    tmad[:,0]-=tmad[0,0]
    return tmad, label

def create_datasets(
        root = 'data/nmnist/n_mnist.hdf5',
        batch_size = 72 ,
        chunk_size_train = 300,
        chunk_size_test = 300,
        ds = 1,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None):

    size = [2, 32//ds, 32//ds]
    print(size)

    if transform_train is None:
        transform_train = Compose([
            CropDims(low_crop=[0,0], high_crop=[32,32], dims=[2,3]),
            Downsample(factor=[dt,1,1,1]),
            ToCountFrame(T = chunk_size_train, size = size),
            ToTensor()])
    if transform_test is None:
        transform_test = Compose([
            CropDims(low_crop=[0,0], high_crop=[32,32], dims=[2,3]),
            Downsample(factor=[dt,1,1,1]),
            ToCountFrame(T = chunk_size_test, size = size),
            ToTensor()])
    if target_transform_train is None:
        target_transform_train =Compose([Repeat(chunk_size_train), toOneHot(10)])
    if target_transform_test is None:
        target_transform_test = Compose([Repeat(chunk_size_test), toOneHot(10)])

    train_ds = NMNISTDataset(root,train=True,
                                 transform = transform_train, 
                                 target_transform = target_transform_train, 
                                 chunk_size = chunk_size_train)

    test_ds = NMNISTDataset(root, transform = transform_test, 
                                 target_transform = target_transform_test, 
                                 train=False,
                                 chunk_size = chunk_size_test)

    return train_ds, test_ds

def create_dataloader(
        root = 'data/nmnist/n_mnist.hdf5',
        batch_size = 72 ,
        chunk_size_train = 300,
        chunk_size_test = 300,
        ds = 1,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        **dl_kwargs):

    train_d, test_d = create_datasets(
        root = 'data/nmnist/n_mnist.hdf5',
        batch_size = batch_size,
        chunk_size_train = chunk_size_train,
        chunk_size_test = chunk_size_test,
        ds = ds,
        dt = dt,
        transform_train = transform_train,
        transform_test = transform_test,
        target_transform_train = target_transform_train,
        target_transform_test = target_transform_test)


    train_dl = torch.utils.data.DataLoader(train_d, shuffle=True, batch_size=batch_size, **dl_kwargs)
    test_dl = torch.utils.data.DataLoader(test_d, shuffle=False, batch_size=batch_size, **dl_kwargs)

    return train_dl, test_dl






#!/bin/python
# -----------------------------------------------------------------------------
# File Name : events_timeslices.py
# Author: Emre Neftci
#
# Creation Date : Thu 16 May 2019 02:13:09 PM PDT
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
# -----------------------------------------------------------------------------


from __future__ import print_function
import bisect
import numpy as np
# from scipy.sparse import coo_matrix as sparse_matrix

def expand_targets(targets, T=500, burnin=0):
    y = np.tile(targets.copy(), [T, 1, 1])
    y[:burnin] = 0
    return y


def one_hot(mbt, num_classes):
    out = np.zeros([mbt.shape[0], num_classes])
    out[np.arange(mbt.shape[0], dtype='int'), mbt.astype('int')] = 1
    return out


def find_first(a, tgt):
    return bisect.bisect_left(a, tgt)


def cast_evs(evs):
    ts = (evs[:, 0] * 1e6).astype('uint64')
    ad = (evs[:, 1:]).astype('uint64')
    return ts, ad

# def get_binary_frame(evs, size = (346,260), ds=1):
#     tr = sparse_matrix((2*evs[:,3]-1,(evs[:,1]//ds,evs[:,2]//ds)), dtype=np.int8, shape=size)
#     return tr.toarray()

def get_subsampled_coordinates(evs, ds_h, ds_w):
    x_coords = evs[:, 1] // ds_w
    y_coords = evs[:, 2] // ds_h
    if x_coords.dtype != np.int:
        x_coords = x_coords.astype(int)
    if y_coords.dtype != np.int:
        y_coords = y_coords.astype(int)
    return x_coords, y_coords


def get_binary_frame_np(arr, evs, ds_w=1, ds_h=1):
    x_coords, y_coords = get_subsampled_coordinates(evs, ds_h, ds_w)
    arr[x_coords, y_coords] = 2 * evs[:, 3] - 1


def get_binary_frame(arr, evs, ds_w=1, ds_h=1):
    x_coords, y_coords = get_subsampled_coordinates(evs, ds_h, ds_w)
    arr[x_coords, y_coords] = 1

def get_slice(times, addrs, start_time, end_time):
    try:
        idx_beg = find_first(times, start_time)
        idx_end = find_first(times[idx_beg:], end_time)+idx_beg
        return times[idx_beg:idx_end]-times[idx_beg], addrs[idx_beg:idx_end]
    except IndexError:
        raise IndexError("Empty batch found")
    
def get_event_slice(times, addrs, start_time, T, size = [128,128], ds = 1, dt = 1000):
    try:
        idx_beg = find_first(times, start_time)
        idx_end = find_first(times[idx_beg:], start_time+T*dt)+idx_beg
        return chunk_evs_pol_dvs(times[idx_beg:idx_end], addrs[idx_beg:idx_end], deltat=dt, chunk_size=T, size = size, ds_w=ds, ds_h=ds)
    except IndexError:
        raise IndexError("Empty batch found")

def get_tmad_slice(times, addrs, start_time, T):
    try:
        idx_beg = find_first(times, start_time)
        idx_end = find_first(times[idx_beg:], start_time+T)+idx_beg
        return np.column_stack([times[idx_beg:idx_end], addrs[idx_beg:idx_end]])
    except IndexError:
        raise IndexError("Empty batch found")

def get_time_surface(evs, invtau=1e-6, size=(346, 260, 2)):
    tr = np.zeros(size, 'int64') - np.inf

    for ev in evs:
        tr[ev[2], ev[1], ev[3]] = ev[0]

    a = np.exp(tr[:, :, 0] * invtau) - np.exp(tr[:, :, 1] * invtau)

    return a


def chunk_evs_dvs(evs, deltat=1000, chunk_size=500, size=[304, 240], ds_w=1, ds_h=1):
    t_start = evs[0, 0]
    ts = range(t_start + chunk_size, t_start + chunk_size * deltat, deltat)
    chunks = np.zeros([len(ts)] + size, dtype='int8')
    idx_start = 0
    idx_end = 0
    for i, t in enumerate(ts):
        idx_end += find_first(evs[idx_end:, 0], t)
        if idx_end > idx_start:
            get_binary_frame_np(chunks[i, ...], evs[idx_start:idx_end], ds_h=ds_h, ds_w=ds_w)
        idx_start = idx_end
    return chunks

def frame_evs(times, addrs, deltat=1000, duration=500, size=[240], downsample = [1]):
    t_start = times[0]
    ts = range(t_start, t_start + duration * deltat, deltat)
    chunks = np.zeros([len(ts)] + size, dtype='int8')
    idx_start = 0
    idx_end = 0
    for i, t in enumerate(ts):
        idx_end += find_first(times[idx_end:], t)
        if idx_end > idx_start:
            ee = addrs[idx_start:idx_end]
            ev = [(ee[:, i] // d).astype(np.int) for i,d in enumerate(downsample)]
            np.add.at(chunks, tuple([i]+ev), 1)
        idx_start = idx_end
    return chunks


def chunk_evs_pol_dvs(times, addrs, deltat=1000, chunk_size=500, size=[2, 304, 240], ds_w=1, ds_h=1):
    t_start = times[0]
    ts = range(t_start, t_start + chunk_size * deltat, deltat)
    chunks = np.zeros([len(ts)] + size, dtype='int8')
    idx_start = 0
    idx_end = 0
    for i, t in enumerate(ts):
        idx_end += find_first(times[idx_end:], t)
        if idx_end > idx_start:
            ee = addrs[idx_start:idx_end]
            pol, x, y = ee[:, 2], (ee[:, 0] // ds_w).astype(np.int), (ee[:, 1] // ds_h).astype(np.int)
            np.add.at(chunks, (i, pol, x, y), 1)
        idx_start = idx_end
    return chunks


if __name__ == "__main__":
    import h5py

    dataset = h5py.File('/home/eneftci_local/Projects/share/data/massiset/massiset_sparse.hdf5', 'r')
    evs = dataset.get('backpack')['data_train'].value
    cevs = chunk_evs(evs, chunk_size=500, deltat=1000, size=[304 // 4, 240 // 4], ds_w=4, ds_h=4)



import struct
import numpy as np
import scipy.misc
import h5py
import glob
import os
import torch.utils.data
from ..events_timeslices import frame_evs, one_hot

SAMPLING_RATE = 200
mapping = { 0 :'Pinch1' ,
            1 :'Pinch2' ,
            2 :'Pinch3' ,
            3 :'Pinch4' ,
            -1 :'none' }
inv_mapping = {'Pinch1': 0,
               'Pinch2': 1,
               'Pinch3': 2,
               'Pinch4': 3,
               'none'  :-1}



def one_hot1d(mbt, num_classes):
    out = np.zeros([num_classes], dtype='float32')
    out[int(mbt)] = 1
    return out

def make_infinite(dataloader):
    while True:
        for x in iter(dataloader):
            yield x
            
class EMGGesturesDataset(torch.utils.data.Dataset):
    nclasses = 11
    
    def __init__(self, evs, labels, ds=[1], size=[64], dt=1000, max_duration=1000):
        super(NTIdigitsDataset).__init__()
        self.evs = evs
        self.labels = labels
        self.ds = ds
        self.dt = dt
        self.size = size
        self.max_duration = max_duration

    def __getitem__(self, key):
        tm = self.evs[key][:,0]
        ad = self.evs[key][:,1:2]
        #T = np.minimum(tm[-1]//self.dt,self.max_duration)
        T = self.max_duration
        sample = frame_evs(tm, ad, duration=T, downsample=[self.ds], size=self.size, deltat=self.dt)
        #learn_mask = np.convolve(np.ones(100)/100,sample.reshape(-1,self.size[0]).sum(axis=1))[:T]>1e-3
        learn_mask = sample.reshape(-1,self.size[0]).sum(axis=1)>0
        targets = np.ones(T)*11
        targets[learn_mask] = self.labels[key]
        return np.array(sample, dtype='float32'), one_hot(targets, self.nclasses+1)[:,:self.nclasses].astype('float32')
    
    def __len__(self):
        return len(self.evs)
        

def create_data(filename = 'data/pinch.hdf5',
                chunk_size_train=10,
                chunk_size_test=20,
                batch_size=50,
                size=[2, 128, 128],
                dt = 1000,
                ds = 1,
                download = True,
                **dl_kwargs):
    
    """
    chunk_size_train: number of samples in the time axis for the training set
    chunk_size_test: number of samples in the time axis for the testing set
    batch_size: batch_size on each iteration
    size: expected input size, list in the format [channels, dimx]
    dt: framing delta t in microseconds
    ds: downsampling factor in the dimx dimension. 1 means no downsampling
    """
    if not os.path.isfile(filename):
        raise Exception("File {} does not exist".format(filename))
    else:
        from torchvision.datasets.utils import download_and_extract_archive
        download_and_extract_archive('https://zenodo.org/record/3194792/files/Pinch.zip?download=1', 'data/', extract_root = 'data/', filename = 'pinch.zip', remove_finished = True)

        with h5py.File(filename, 'w') as f:
            f.clear()
            data = f.create_group('data')
            extras = f.create_group('extras')

            labels = f.create_group('labels')


            emg_data = glob.glob('data/Pinch/*emg.npy')

            lab_data = glob.glob('data/Pinch/*ann.npy')

            key=0
            for ef in emg_data:
                ef_data = np.load(ef)
                ef_labels = np.load(ef[:-8]+"_ann.npy")
                ef_labels_idx = np.zeros([ef_labels.shape[0]])
                for k,v in inv_mapping.items():
                    ef_labels_idx[ef_labels==k]=v

                ef_time = np.linspace(0, ef_data.shape[0]*1000000/SAMPLING_RATE, ef_data.shape[0],dtype='uint32')
                dev  = data.create_group(str(key))
                dtgt = labels.create_group(str(key))
                dset_dt = dev.create_dataset('time', data = ef_time, dtype=np.uint32)
                dset_da = dev.create_dataset('data', data = ef_data, dtype=np.int32)
                dset_l  = dtgt.create_dataset('class', data=ef_labels_idx, dtype=np.int8)
                key = key + 1



            dset_dt = events.create_dataset('time',  [], dtype=np.uint32)
            dset_da = events.create_dataset('data',  [], dtype=np.uint16)
            dset_l  = labels.create_dataset('class', [], dtype=np.uint32)








import struct
import time
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from ..neuromorphic_dataset import NeuromorphicDataset 
from ..events_timeslices import *
from ..transforms import *
import os

mapping = range(35)

def one_hot1d(mbt, num_classes):
    out = np.zeros([num_classes], dtype='float32')
    out[int(mbt)] = 1
    return out

def create_events_hdf5(directory, hdf5_filename):
    train_evs, train_labels_isolated = load_shd_hdf5(directory+'/ssc_train.h5')
    test_evs, test_labels_isolated = load_shd_hdf5(directory+'/ssc_test.h5')
    border = len(train_labels_isolated)

    tmad = train_evs + test_evs
    labels = train_labels_isolated + test_labels_isolated 
    test_keys = []
    train_keys = []

    with h5py.File(hdf5_filename, 'w') as f:
        f.clear()
        key = 0
        metas = []
        data_grp = f.create_group('data')
        extra_grp = f.create_group('extra')
        for i,data in enumerate(tmad):
            times = data[:,0]
            addrs = data[:,1:]
            label = labels[i]
            out = []
            istrain = i<border
            if istrain: 
                train_keys.append(key)
            else:
                test_keys.append(key)
            metas.append({'key':str(key), 'training sample':istrain}) 
            subgrp = data_grp.create_group(str(key))
            tm_dset = subgrp.create_dataset('times' , data=times, dtype=np.uint32)
            ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype=np.uint16)
            lbl_dset= subgrp.create_dataset('labels', data=label, dtype=np.uint8)
            subgrp.attrs['meta_info']= str(metas[-1])
            assert label in mapping
            key += 1
        extra_grp.create_dataset('train_keys', data=train_keys)
        extra_grp.create_dataset('test_keys', data=test_keys)
        extra_grp.attrs['N'] = len(train_keys) + len(test_keys)
        extra_grp.attrs['Ntrain'] = len(train_keys)
        extra_grp.attrs['Ntest'] = len(test_keys)

def load_shd_hdf5(filename, train=True):
    with h5py.File(filename, 'r', swmr=True, libver="latest") as f:
        evs = []
        labels = []
        for i,tl in enumerate(f['labels']):
            label_ = tl
            digit = tl
            digit=int(digit)
            labels.append(digit)
            tm = np.int32(f['spikes']['times'][i][:]*1e6)
            ad = np.int32(f['spikes']['units'][i][:])
            evs.append(np.column_stack([tm,ad]))
        return evs, labels

            
class SSCDataset(NeuromorphicDataset):
    resources_url = [['https://compneuro.net/datasets/ssc_test.h5.gz',None, 'ssc_test.h5.gz'], ['https://compneuro.net/datasets/ssc_train.h5.gz',None, 'ssc_train.h5.gz']]
    directory = 'data/shd/'
    resources_local = [directory+'/ssc_train.hdf5',directory+'/ssc_test.hdf5']


    def __init__(
            self, 
            root,
            train=True,
            transform=None,
            target_transform=None,
            download_and_create=True,
            chunk_size = 500):

        self.n = 0
        self.download_and_create = download_and_create
        self.root = root
        self.train = train 
        self.chunk_size = chunk_size

        super(SSCDataset, self).__init__(
                root,
                transform=transform,
                target_transform=target_transform )
        
        with h5py.File(root, 'r', swmr=True, libver="latest") as f:
            if train:
                self.n = f['extra'].attrs['Ntrain']
                self.keys = f['extra']['train_keys']
            else:
                self.n = f['extra'].attrs['Ntest']
                self.keys = f['extra']['test_keys']

    def download(self):
        isexisting = super(SSCDataset, self).download()

    def create_hdf5(self):
        create_events_hdf5(self.directory, self.root)


    def __len__(self):
        return self.n
        
    def __getitem__(self, key):
        #Important to open and close in getitem to enable num_workers>0
        with h5py.File(self.root, 'r', swmr=True, libver="latest") as f:
            if not self.train:
                key = key + f['extra'].attrs['Ntrain']
            data, target = sample(
                    f,
                    key,
                    T = self.chunk_size,
                    shuffle=self.train)

        if self.transform is not None:
            data = self.transform(data)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return data, target

def sample(hdf5_file,
        key,
        T = 500,
        shuffle = False):
    dset = hdf5_file['data'][str(key)]
    label = dset['labels'][()]
    tend = dset['times'][-1] 
    start_time = 0

    tmad = get_tmad_slice(dset['times'][()], dset['addrs'][()], start_time, T*1000)
    tmad[:,0]-=tmad[0,0]
    return tmad, label

def create_dataloader(
        root = 'data/shd/shd.hdf5',
        batch_size = 72 ,
        chunk_size_train = 1000,
        chunk_size_test = 1000,
        ds = 1,
        dt = 500,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        **dl_kwargs):

    size = [700//ds, 1, 1]

    if transform_train is None:
        transform_train = Compose([
            Downsample(factor=[dt,ds]),
            ToChannelHeightWidth(),
            ToCountFrame(T = chunk_size_train, size = size),
            ToTensor()
            ])
    if transform_test is None:
        transform_test = Compose([
            Downsample(factor=[dt,ds]),
            ToChannelHeightWidth(),
            ToCountFrame(T = chunk_size_test, size = size),
            ToTensor()
            ])
    if target_transform_train is None:
        target_transform_train =Compose([Repeat(chunk_size_train), toOneHot(len(mapping))])
    if target_transform_test is None:
        target_transform_test = Compose([Repeat(chunk_size_test), toOneHot(len(mapping))])

    collate_fn = None
    #collate_fn = lambda x: zip(*[x[i] for i in range(len(x))])

    train_d = SSCDataset(root,train=True,
                                 transform = transform_train, 
                                 target_transform = target_transform_train, 
                                 chunk_size = chunk_size_train)

    train_dl = torch.utils.data.DataLoader(train_d, shuffle=True, batch_size=batch_size, collate_fn = collate_fn, **dl_kwargs)

    test_d = SSCDataset(root, transform = transform_test, 
                                 target_transform = target_transform_test, 
                                 train=False,
                                 chunk_size = chunk_size_test)

    test_dl = torch.utils.data.DataLoader(test_d, shuffle=False, batch_size=batch_size, collate_fn = collate_fn, **dl_kwargs)
                                 

    return train_dl, test_dl



import struct
import time
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from ..neuromorphic_dataset import NeuromorphicDataset 
from ..events_timeslices import *
from ..transforms import *
import os

mapping = { 0 :'0',
            1 :'1',
            2 :'2',
            3 :'3',
            4 :'4',
            5 :'5',
            6 :'6',
            7 :'7',
            8 :'8',
            9 :'9',
            10: '10'}

def one_hot1d(mbt, num_classes):
    out = np.zeros([num_classes], dtype='float32')
    out[int(mbt)] = 1
    return out

def create_events_hdf5(directory, hdf5_filename):
    train_evs, train_labels_isolated = load_tidigit_hdf5(directory+'/n-tidigits.hdf5', train=True)
    test_evs, test_labels_isolated = load_tidigit_hdf5(directory+'/n-tidigits.hdf5', train=False)
    border = len(train_labels_isolated)

    tmad = train_evs + test_evs
    labels = train_labels_isolated + test_labels_isolated 
    test_keys = []
    train_keys = []

    with h5py.File(hdf5_filename, 'w') as f:
        f.clear()
        key = 0
        metas = []
        data_grp = f.create_group('data')
        extra_grp = f.create_group('extra')
        for i,data in enumerate(tmad):
            times = data[:,0]
            addrs = data[:,1:]
            label = labels[i]
            out = []
            istrain = i<border
            if istrain: 
                train_keys.append(key)
            else:
                test_keys.append(key)
            metas.append({'key':str(key), 'training sample':istrain}) 
            subgrp = data_grp.create_group(str(key))
            tm_dset = subgrp.create_dataset('times' , data=times, dtype=np.uint32)
            ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype=np.uint8)
            lbl_dset= subgrp.create_dataset('labels', data=label, dtype=np.uint8)
            subgrp.attrs['meta_info']= str(metas[-1])
            assert label in mapping
            key += 1
        extra_grp.create_dataset('train_keys', data=train_keys)
        extra_grp.create_dataset('test_keys', data=test_keys)
        extra_grp.attrs['N'] = len(train_keys) + len(test_keys)
        extra_grp.attrs['Ntrain'] = len(train_keys)
        extra_grp.attrs['Ntest'] = len(test_keys)

def load_tidigit_hdf5(filename, train=True):
    with h5py.File(filename, 'r', swmr=True, libver="latest") as f:
        if train:
            train_evs = []
            train_labels_isolated = []
            for tl in f['train_labels']:
                label_ = tl.decode()
                label_s = label_.split('-')
                if len(label_s[-1])==1:
                    digit = label_s[-1]
                    if digit is 'o':
                        digit=10
                    if digit is 'z':
                        digit=0
                    else:
                        digit=int(digit)
                    train_labels_isolated.append(digit)
                    tm = np.int32(f['train_timestamps'][label_][:]*1e6)
                    ad = np.int32(f['train_addresses'][label_][:])
                    train_evs.append(np.column_stack([tm,ad]))
            return train_evs, train_labels_isolated

        else:
            test_evs = []
            test_labels_isolated  = []
            for tl in f['test_labels']:
                label_ = tl.decode()
                label_s = label_.split('-')
                if len(label_s[-1])==1:
                    digit = label_s[-1]
                    if digit is 'o':
                        digit=10
                    if digit is 'z':
                        digit=0
                    else:
                        digit=int(digit)
                    test_labels_isolated.append(digit)
                    tm = np.int32(f['test_timestamps'][label_][:]*1e6)
                    ad = np.int32(f['test_addresses'][label_].value)
                    test_evs.append(np.column_stack([tm,ad]))
            
            return test_evs, test_labels_isolated
            
#class NTIdigitsDataset(torch.utils.data.Dataset):
#    nclasses = 11
#    
#    def __init__(self, 
#            filename,
#            train=True,
#            ds=[1], #transform
#            size=[64], #transform
#            dt=1000, #transform
#            max_duration=1000, #transform
#            download=False
#            ):
#        super(NTIdigitsDataset).__init__()
#        if download:
#            raise NotImplementedError()
#        evs, labels = load_tidigit_hdf5(filename, train=train)
#        self.evs = evs
#        self.labels = labels
#        self.ds = ds
#        self.dt = dt
#        self.size = size
#        self.max_duration = max_duration
#
#    def __getitem__(self, key):
#        tm = self.evs[key][:,0]
#        ad = self.evs[key][:,1:2]
#        #T = np.minimum(tm[-1]//self.dt,self.max_duration)
#        T = self.max_duration
#        sample = frame_evs(tm, ad, duration=T, downsample=[self.ds], size=self.size, deltat=self.dt)
#        #learn_mask = np.convolve(np.ones(100)/100,sample.reshape(-1,self.size[0]).sum(axis=1))[:T]>1e-3
#        learn_mask = sample.reshape(-1,self.size[0]).sum(axis=1)>0
#        targets = np.ones(T)*11
#        targets[learn_mask] = self.labels[key]
#        return np.array(sample, dtype='float32'), one_hot(targets, self.nclasses+1)[:,:self.nclasses].astype('float32')
#    
#    def __len__(self):
#        return len(self.evs)
        
#
#def create_data(filename = 'data/tidigits/n-tidigits.hdf5',
#                chunk_size_train=10,
#                chunk_size_test=20,
#                batch_size=50,
#                size=[2, 128, 128],
#                dt = 1000,
#                ds = 1,
#                **dl_kwargs):
#    
#    """
#    chunk_size_train: number of samples in the time axis for the training set
#    chunk_size_test: number of samples in the time axis for the testing set
#    batch_size: batch_size on each iteration
#    size: expected input size, list in the format [channels, dimx]
#    dt: framing delta t in microseconds
#    ds: downsampling factor in the dimx dimension. 1 means no downsampling
#    """
#    if not os.path.isfile(filename):
#        raise Exception("File {} does not exist".format(filename))
#
#    train_d = NTIdigitsDataset(filename, train=True, ds = ds, size = size, dt = dt, max_duration = chunk_size_train)
#    train_dl = torch.utils.data.DataLoader(train_d,
#                                           batch_size=batch_size,
#                                           shuffle=True, **dl_kwargs)
#    
#    test_d   = NTIdigitsDataset(filename, train=False, ds = ds, size = size, dt = dt, max_duration = chunk_size_test)
#    test_dl = torch.utils.data.DataLoader( test_d,
#                                           batch_size=batch_size,
#                                           shuffle=True, **dl_kwargs)
#
#    return train_dl, test_dl

class NTIdigitsDataset(NeuromorphicDataset):
    resources_url = [['https://www.dropbox.com/s/vfwwrhlyzkax4a2/n-tidigits.hdf5?dl=1',None, 'n-tidigits.hdf5']]
    directory = 'data/tidigits/'
    resources_local = [directory+'/n-tidigits.hdf5']


    def __init__(
            self, 
            root,
            train=True,
            transform=None,
            target_transform=None,
            download_and_create=True,
            chunk_size = 500):

        self.n = 0
        self.download_and_create = download_and_create
        self.root = root
        self.train = train 
        self.chunk_size = chunk_size

        super(NTIdigitsDataset, self).__init__(
                root,
                transform=transform,
                target_transform=target_transform )
        
        with h5py.File(root, 'r', swmr=True, libver="latest") as f:
            if train:
                self.n = f['extra'].attrs['Ntrain']
                self.keys = f['extra']['train_keys']
            else:
                self.n = f['extra'].attrs['Ntest']
                self.keys = f['extra']['test_keys']

    def download(self):
        isexisting = super(NTIdigitsDataset, self).download()

    def create_hdf5(self):
        create_events_hdf5(self.directory, self.root)


    def __len__(self):
        return self.n
        
    def __getitem__(self, key):
        #Important to open and close in getitem to enable num_workers>0
        with h5py.File(self.root, 'r', swmr=True, libver="latest") as f:
            if not self.train:
                key = key + f['extra'].attrs['Ntrain']
            data, target = sample(
                    f,
                    key,
                    T = self.chunk_size,
                    shuffle=self.train)

        if self.transform is not None:
            data = self.transform(data)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return data, target

def sample(hdf5_file,
        key,
        T = 500,
        shuffle = False):
    dset = hdf5_file['data'][str(key)]
    label = dset['labels'][()]
    tend = dset['times'][-1] 
    start_time = 0

    tmad = get_tmad_slice(dset['times'][()], dset['addrs'][()], start_time, T*1000)
    tmad[:,0]-=tmad[0,0]
    return tmad, label

def create_dataloader(
        root = 'data/tidigits/ntidigits_isolated.hdf5',
        batch_size = 72 ,
        chunk_size_train = 1000,
        chunk_size_test = 1000,
        ds = 1,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        **dl_kwargs):

    size = [64//ds[0], 1, 1]

    if transform_train is None:
        transform_train = Compose([
            Downsample(factor=[dt,ds[0]]),
            ToChannelHeightWidth(),
            ToCountFrame(T = chunk_size_train, size = size),
            ToTensor()])
    if transform_test is None:
        transform_test = Compose([
            Downsample(factor=[dt,ds[0]]),
            ToChannelHeightWidth(),
            ToCountFrame(T = chunk_size_test, size = size),
            ToTensor()])
    if target_transform_train is None:
        target_transform_train =Compose([Repeat(chunk_size_train), toOneHot(len(mapping))])
    if target_transform_test is None:
        target_transform_test = Compose([Repeat(chunk_size_test), toOneHot(len(mapping))])

    train_d = NTIdigitsDataset(root,train=True,
                                 transform = transform_train, 
                                 target_transform = target_transform_train, 
                                 chunk_size = chunk_size_train)

    train_dl = torch.utils.data.DataLoader(train_d, shuffle=True, batch_size=batch_size, **dl_kwargs)

    test_d = NTIdigitsDataset(root, transform = transform_test, 
                                 target_transform = target_transform_test, 
                                 train=False,
                                 chunk_size = chunk_size_test)

    test_dl = torch.utils.data.DataLoader(test_d, shuffle=False, batch_size=batch_size, **dl_kwargs)

    return train_dl, test_dl



"""
Torch Neuromorphic
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "torchneuromorphic",
    version = "0.2",
    author = "Emre Neftci",
    author_email = "eneftci@uci.edu",
    description = ("Dataset loaders for pytorch"),
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    download_url='https://github.com/nmi-lab/torchneuromorphic/archive/0.2.tar.gz',
    long_description=long_description,
    license='Apache License 2.0',
    install_requires=[
        "torchvision>=0.4.1",
        "torch>=1.1.0",
        "scipy>=1.0",
        "h5py",
        "pandas",
    ]
)
